@@ -1,6 +1,7 @@
 """Kazoo Zookeeper Client"""
 import inspect
 import logging
+import random
 from collections import namedtuple
 from functools import partial
 from os.path import split
@@ -24,6 +25,13 @@ log = logging.getLogger(__name__)
 
 ZK_OPEN_ACL_UNSAFE = {"perms": zookeeper.PERM_ALL, "scheme": "world",
                        "id": "anyone"}
+ZK_STATES = {}
+ZK_TYPES = {}
+for name, val in zookeeper.__dict__.items():
+    if name.endswith('_STATE'):
+        ZK_STATES[val] = name
+    if name.endswith('EVENT'):
+        ZK_TYPES[val] = name
 
 
 ## Client State and Event objects
@@ -339,6 +347,11 @@ class KazooClient(object):
                         :class:`~kazoo.interfaces.IHandler` interface
                         for callback handling.
         :param default_acl: A default ACL used on node creation.
+
+
+        Retry parameters will be used for connection establishment attempts
+        and reconnects.
+
         """
         from kazoo.recipe.barrier import Barrier
         from kazoo.recipe.barrier import DoubleBarrier
@@ -393,6 +406,12 @@ class KazooClient(object):
             sleep_func=self.handler.sleep_func
         )
 
+        self.max_retries = max_retries
+        self.retry_delay = retry_delay
+        self.retry_backoff = retry_backoff
+        self.retry_jitter = int(retry_jitter * 100)
+        self._connection_attempts = None
+
         # Curator like simplified state tracking, and listeners for state
         # transitions
         self.state = KazooState.LOST
@@ -509,11 +528,18 @@ class KazooClient(object):
         # remove itself
         for listener in list(self.state_listeners):
             try:
-                listener(state)
+                remove = listener(state)
+                if remove is True:
+                    self.remove_listener(listener)
             except Exception:
                 log.exception("Error in connection state listener")
 
     def _session_callback(self, handle, type, state, path):
+        log.debug("Client Instance: %s, Handle: %s, Type: %s, State: %s"
+                  ", Path: %s", id(self), handle, ZK_TYPES.get(type, type),
+                  ZK_STATES.get(state, state),
+                  path)
+
         if type != EventType.SESSION:
             return
 
@@ -530,15 +556,27 @@ class KazooClient(object):
             return
 
         if state == KeeperState.CONNECTED:
+            log.info("Zookeeper connection established")
             self._live.set()
+            self._connection_attempts = None
+
+            # Clear the client id when we successfully connect
+            self._provided_client_id = None
+
             self._make_state_change(KazooState.CONNECTED)
         elif state in (KeeperState.EXPIRED_SESSION,
                        KeeperState.AUTH_FAILED):
+            log.info("Zookeeper session lost, state: %s", state)
             self._live.clear()
             self._make_state_change(KazooState.LOST)
             self._handle = None
-            self.connect_async()
+
+            # This session callback already runs in the handler so
+            # its safe to spawn the start_async call so this function
+            # can return immediately
+            self.handler.spawn(self._reconnect)
         else:
+            log.info("Zookeeper connection lost")
             # Connection lost
             self._live.clear()
             self._make_state_change(KazooState.SUSPENDED)
@@ -581,6 +619,8 @@ class KazooClient(object):
         self.handler.start()
 
         cb = self._wrap_session_callback(self._session_callback)
+        self._connection_attempts = 1
+        self._connection_delay = self.retry_delay
         if self._provided_client_id:
             self._handle = self.zookeeper.init(self._hosts, cb, self._timeout,
                 self._provided_client_id)
@@ -589,6 +629,20 @@ class KazooClient(object):
         return self._live
     connect_async = start_async
 
+    def _reconnect(self):
+        """Reconnect to Zookeeper, staggering connection attempts"""
+        if not self._connection_attempts:
+            self._connection_attempts = 1
+            self._connection_delay = self.retry_delay
+        else:
+            if self._connection_attempts == self.max_retries:
+                raise self.handler.timeout_exception("Time out reconnecting")
+            self._connection_attempts += 1
+            jitter = random.randint(0, self.retry_jitter) / 100.0
+            self.handler.sleep_func(self._connection_delay + jitter)
+            self._connection_delay *= self.retry_delay
+        self.start_async()
+
     def start(self, timeout=15):
         """Initiate connection to ZK
 
